(module board mzscheme
  
  ;; ---------------------------------------- interface ----------------------------------------
  
  ;; board-width : board -> num
  ;; board-height : board -> num
  ;; board-ref : board num num -> (U 'plain 'water 'wall 'home)
  ;;   first num is x-coord, second num is y-coord (both are 1-based indexing)
  ;;   returns one of those four symbols
  ;; read-board-from-file : str -> board
  
  ;; -------------------------------------------------------------------------------------------
  
  (require (lib "match.ss")
           "utils.ss")
  
  (provide read-board
           read-board-from-file
           board-width
           board-height
           board-ref)
  
  (define-struct board (width height tiles))

  (define (decode-char c)
    (case c
      [(#\.) 'plain]
      [(#\~) 'water]
      [(#\#) 'wall]
      [(#\@) 'home]))

  ;; read-board : input-port -> board
  (define (read-board in)
    (match-let ([(width height) (map string->number (read-list in))])
      (let ([tiles (make-vector height)])
        (let loop ([l 0])
          (when (< l height)
            (vector-set! tiles l (list->vector (map decode-char (string->list (read+log-line in)))))
            (loop (+ l 1))))
        (make-board width height tiles))))

  ;; read-board-from-file : str -> board
  (define (read-board-from-file filename)
    (call-with-input-file filename read-board))
  
  ;; board-ref : board num num -> (U 'plain 'water 'wall 'home)
  ;; 1-based indexing
  (define (board-ref board x y)
    (if (or (<= x 0) (<= y 0) (> x (board-width board)) (> y (board-height board)))
        'wall
        (vector-ref (vector-ref (board-tiles board) (- y 1)) (- x 1))))
  )