(module viz mzscheme
  
  (require (lib "class.ss")
           (lib "match.ss")
           (lib "mred.ss" "mred")
           "board.ss"
           "types.ss")
  
  (provide (all-defined))
  
  (define viz%
    (class object%
      
      (init-field board
                  my-id
                  (scale 30)
                  (on-close (lambda () (exit))))
      
      (define frame (instantiate frame%
                      ((string-append "Player " (number->string my-id)))))

      (define canvas (instantiate canvas% (frame)
                       (min-width (* scale (board-width board)))
                       (min-height (* scale (board-height board)))))
      ;      (define canvas (instantiate canvas% (frame) 
      ;      (paint-callback (lambda (canvas dc) (draw-face dc)))))
      
      (define score-message (instantiate message% ("Score: 0" frame)
                              (min-width 400)))
      
      (define update-message (instantiate message% ("No updates yet" frame)
                               (min-width 400)))
      
      (define close-button (instantiate button% () (label "Close") (parent frame)
                             ;; Callback procedure for a button click
                             (callback (lambda (button event)
                                         (on-close)))))
      
      ;; Make some pens and brushes
      (define no-pen (instantiate pen% ("Black" 1 'transparent)))
      (define plain-brush (instantiate brush% ("LawnGreen" 'solid)))
      (define water-brush (instantiate brush% ("Aqua" 'solid)))
      (define wall-brush (instantiate brush% ("Firebrick" 'solid)))
      (define home-brush (instantiate brush% ("Gold" 'solid)))
      
      ;; u-actions->string : (listof u-action) -> str
      (define (u-actions->string u)
        (if (null? u)
            ""
            (string-append (match (car u)
                             [($ u-move dir) (string-append "moved " (symbol->string dir))]
                             [($ u-pick id) (string-append "picked up " (number->string id))]
                             [($ u-drop id) (string-append "dropped up " (number->string id))]
                             [($ u-loc x y) (string-append "appeared at " (number->string x) ":" (number->string y))])
                           ", "
                           (u-actions->string (cdr u)))))
        
      ;; my-update-string (listof update) -> str
      (define (my-update-string updates)
        (if (null? updates)
            "dead"
            (if (equal? my-id (update-id (car updates)))
                (string-append "Last round: " (u-actions->string (update-u-actions (car updates))) "stop")
                (my-update-string (cdr updates)))))
      
      (define/public (draw-board players updates my-score)
        (let ([dc (send canvas get-dc)])
          (send dc set-pen no-pen)
          (let loop-y ([y 1])
            (when (<= y (board-height board))
              (let loop-x ([x 1])
                (when (<= x (board-width board))
                  (case (board-ref board x y)
                    [(plain) (send dc set-brush plain-brush)]
                    [(water) (send dc set-brush water-brush)]
                    [(wall) (send dc set-brush wall-brush)]
                    [(home) (send dc set-brush home-brush)])
                  (send dc draw-rectangle (* (- x 1) scale) (* (- (board-height board) y) scale) scale scale)
                  (loop-x (+ x 1))))
              (loop-y (+ y 1))))
          (for-each (lambda (p)
                      (send dc draw-text
                            (number->string (player-id p))
                            (* (- (player-x p) 1) scale)
                            (* (- (board-height board) (player-y p)) scale)
                            #f
                            0))
                    players))
        (send score-message set-label (string-append "Score: " (number->string my-score)))
        (send update-message set-label (my-update-string updates))
        (send frame show #t))
      
      (super-instantiate ())))
  )
