(module assert mzscheme
  
  ;; assert: does roughly what you'd expect: evaluates its argument, and if
  ;; arg returns non-false value, continues silently.  Otherwise, abort with
  ;; assorted useful info.  If you are asserting the negation of something,
  ;; use (assert not (...)) instead of (assert (not (...))).
  ;; CAVEAT PROGRAMMOR: Currently, (assert (f x ...)) will evaluate x ... twice
  ;; in the case of assertion failure.  This does not affect program correctness,
  ;; but if x ... has side effects, error output may be incorrect.  On the other 
  ;; hand, using a side-effecty function in an assertion is probably a REALLY BAD
  ;; IDEA anyway, so I'm not too worried.
  (define-syntax (assert stx)
    (let ([handle-assert 
	   (lambda (src-stx positive stx)
	     (let ([src (syntax-source src-stx)]
		   [line (syntax-line src-stx)])
	       (with-syntax ([pfx (lambda (form)
				    (printf "Assertion failed at ~a~a: ~a~s~n"
					    src
					    (if line (format ":~a" line) "")
					    (if positive "" "not ")
					    form))]
			     [sfx (lambda () (error "Aborting execution"))]
			     [id-or-not (if positive (lambda (x) x) not)]
			     [print-arg (lambda (x y)
					  (printf "Value of ~s is ~s~n" x y))])
			    (syntax-case stx (assert)
			      [(assert (a b ...))
			       (syntax (unless (id-or-not (a b ...))
					 (pfx '(a b ...))
					 (let ([eargs (list b ...)])  ; b is evaluated second time
					   (for-each print-arg '(b ...) eargs))
					 (sfx)))]
			      [(assert a)
			       (syntax
				(unless (id-or-not a) (pfx 'a) (print-arg 'a a) (sfx)))]))))])
      (syntax-case stx (assert not)
	[(assert not x) (handle-assert stx #f (syntax (assert x)))]
	[(assert x)     (handle-assert stx #t stx)]))))
