(module cqueue mzscheme
  
  (require "base-gm.ss")
  (provide make-cqueue
	   cqueue-add-last!
	   cqueue-add-first!
	   cqueue-first
	   cqueue-last
	   cqueue-size
	   cqueue-empty?
	   cqueue-remove-first!
	   cqueue-clear!
	   cqueue-map
	   cqueue-for-each
	   cqueue-tostring )
  
  (define-struct t (size head tail))
  
  (define (cqueue-tostring q)
    (format "<cq: sz ~a ~a>" (t-size q) (cdr (t-head q))))
  
  (define debug-f (make-debug cqueue-tostring))
  
  (define (make-cqueue) 
    (let ((dummy (cons empty empty)))
      (make-t 0 dummy dummy)))
  
  (define (cqueue-add-last! q item)
    (let ((cell (cons item empty)))
      (set-t-size! q (+ (t-size q) 1))
      (set-cdr! (t-tail q) cell)
      (set-t-tail! q cell)))
  
  (define (cqueue-add-first! q item)
    (let ((cell (cons item (cdr (t-head q)))))
      (set-t-size! q (+ (t-size q) 1))
      (set-cdr! (t-head q) cell)))
  
  (define cqueue-size t-size)
  
  (define (cqueue-empty? q)
    (= (t-size q) 0))
  
  (define (cqueue-first q)
    (cadr (t-head q)))

  (define (cqueue-last q)
    (car (t-tail q)))

  (define (cqueue-remove-first! q)
    (if (cqueue-empty? q) (error "cqueue-remove-last!: empty")
	(let ((rtn (cdr (t-head q))))
	  (set-t-head! q (cdr (t-head q)))
	  (set-t-size! q (- (t-size q) 1))
	  (car rtn))))
  
  (define (cqueue-clear! q)
    (set-cdr! (t-head q) empty)
    (set-t-tail! q (t-head q))
    (set-t-size! q 0))
  
  (define (cqueue-map fn q)
    (map fn (cdr (t-head q))))

  (define (cqueue-for-each fn q)
    (for-each fn (cdr (t-head q))))

  (define (test)
    (define q (make-cqueue))
    (for-each (lambda (i) 
		(cqueue-add-last! q i)
		(debug-f q)
		(cqueue-add-first! q i)
		(debug-f q)
		)
	      '(1 2 3 4 5 6))
    
    (debug-f "R" (cqueue-remove-first! q) q)
    (debug-f "R" (cqueue-remove-first! q) q)
    (debug-f "R" (cqueue-remove-first! q) q)
    (debug-f "R" (cqueue-remove-first! q) q)
    (cqueue-clear! q)
    (debug-f q)
    (for-each (lambda (i) 
		(cqueue-add-last! q i)
		(debug-f q)
		(cqueue-add-first! q i)
		(debug-f q)
		)
	      '(1 2 3 4 5 6)))
;  (test)
)

