(module minimax mzscheme

  (require "base-gm.ss" "physics.ss" "../rob/draw-util.scm")
  (provide minimax-move minimax eval-max-val)

  (define-struct val (min-v min-pls max-v max-pls my-move))

  (define (val-tostring val)
    (if (not (val? val)) false
	(format "<val ~a ~a ~a ~a ~a>" 
		(val-min-v val)
		(to-string (val-min-pls val) player-tostring)
		(val-max-v val)
		(to-string (val-max-pls val) player-tostring)
		(val-my-move val))))

  (define debug-f (make-debug player-tostring val-tostring))

  (define (zip l1 l2)
    (cond ((and (empty? l1) (empty? l2)) empty)
	  ((or (empty? l1) (empty? l2)) (error "zip: sizeways"))
	  (true (cons (cons (car l1) (car l2)) (zip (cdr l1) (cdr l2))))))

  (define (unzip lst)
    (if (empty? lst) (cons empty empty)
	(let ((a (caar lst))
	      (b (cdar lst))
	      (sub (unzip (cdr lst))))
	  (cons (cons a (car sub)) (cons b (cdr sub))))))



  (define (fold-all-moves board players init fn)
    (let* ((player-vec (list->vector players))
	   (player-l (vector-length player-vec)))

      (let loop ((lst empty) (lst-l 0) (acc init))
	(if (= lst-l player-l) (fn lst acc)
	    (foldl (lambda (dir acc)
		     (loop (cons dir lst) (+ lst-l 1) acc))
		   acc 
		   (player-possible-moves 
		    board (vector-ref player-vec (- player-l lst-l 1))))))))


  (define (fold-permutations target n init fn)
    (let loop ((i 0) (acc init))
      (if (= i n) acc
	  (loop (+ i 1)
		(fn (quicksort target (lambda (a b) (= (random 2) 0))) acc)))))



  (define (fold-the-future board me others
			   compound-me compound-adv compound-perm
			   compound-me-init compound-adv-init compound-perm-init
			   perm-l fn)

    (define players (cons me others))
    (define (l-foldl init lst fn) (foldl fn init lst))
    (define f-moves fold-all-moves)
    (define f-perms fold-permutations)
    (define (compound-me-l my-move acc item) (compound-me my-move item acc))
    (define (compound-adv-l acc item) (compound-adv item acc))
    (define (compound-perm-l n acc item) (compound-perm n item acc))
    
    (l-foldl (compound-me-init) (player-possible-moves board me)
	     (lambda (my-move acc)
	       (compound-me-l
		my-move
		acc
		(f-moves board others (compound-adv-init)
			 (lambda (adv-moves acc)
			   (let ((zipped (zip (cons me others)
					      (cons my-move adv-moves))))

			     (compound-adv-l
			      acc
			      (f-perms zipped perm-l (compound-perm-init)
				       (lambda (perm acc)
					 (let* ((unziped (unzip perm))
						(pls (car unziped))
						(mvs (cdr unziped)))

					   (compound-perm-l
					    perm-l acc
					    (fn (physics-apply-move-sequence board pls mvs))))))))))))))


  (define (player-distance a b)
    (+ (abs (- (player-x a) (player-x b)))
       (abs (- (player-y a) (player-y b)))))


  (define eval-distance-away-adv-mult -2)
  (define eval-kill-value 15)
  (define eval-max-val (* 1000 1000))
  (define eval-perm-l-mult 2)
  (define eval-perm-l 10)
  (define cache-dept-bonus 0)

  (define (val-min! a b)
    (if (and a (< (val-min-v a) (val-min-v b)))
	(begin (set-val-min-v! b (val-min-v a))
	       (set-val-min-pls! b (val-min-pls a))
	       (set-val-max-v! b (val-max-v a))
	       (set-val-max-pls! b (val-max-pls a))
	       true)
	false))

  (define (val-max! a b)
    (when (and a (> (val-max-v a) (val-max-v b)))
      (set-val-min-v! b (val-min-v a))
      (set-val-min-pls! b (val-min-pls a))
      (set-val-max-v! b (val-max-v a))
      (set-val-max-pls! b (val-max-pls a))))

  (define (val-minmax! a b)
    (when (and a (< (val-min-v a) (val-min-v b)))
      (set-val-min-v! b (val-min-v a))
      (set-val-min-pls! b (val-min-pls a)))

    (when (and a (> (val-max-v a) (val-max-v b)))
      (set-val-max-v! b (val-max-v a))
      (set-val-max-pls! b (val-max-pls a))))

  (define (compound-me-init) (make-val eval-max-val empty (- eval-max-val) empty false))
  (define (compound-me my-move val acc)
    (if (val-min! val acc)
	(set-val-my-move! acc my-move))
    acc)

  (define (compound-adv-init) (make-val eval-max-val empty (- eval-max-val) empty false))
  (define (compound-adv val acc)
    (val-max! val acc) 
    acc)

  (define (compound-perm-init) (make-val eval-max-val empty (- eval-max-val) empty false))
  (define (compound-perm perm-l val acc)
    (val-minmax! val acc)
    acc)
  
  

  ;; TODO
  ;;
  ;; biding hints, amount, before/after, rushed
  ;; outside cache with clusters
  ;;
  ;;

  (define (eval-local board cell-map me players)
    (let ((val (+ (cell-map (player-x me) (player-y me))
		  (* eval-distance-away-adv-mult
		     (foldl (lambda (adv acc) (+ acc (player-distance me adv))) 0 players))
		  (* eval-kill-value (length players)))))
      (make-val val players val players false)))

  (define (sub-deadline dept deadline-node-count branches branch-factor)
    (let* ((nodes-left (- deadline-node-count eval-node-count))
	   (branches-left (if (< branches branch-factor) 
			      (- branch-factor branches) 
			      branch-factor))
	   (per-branch (/ nodes-left branches-left))
	   (rtn (+ eval-node-count per-branch)))
      ;;(if (= dept 0) (debug "deadline: " per-branch))
      rtn))

  (define eval-branch-factor 20)
  (define eval-node-per-msec 5.0)
  (define eval-node-count 0)
  (define eval-next-average 0)
  (define eval-time-started 0)

  (define (eval-may-recurse board cell-map my-id players dept-to-go cache deadline deadline-node-count)
    (define (is-me p) (if (= (player-id p) my-id) p false))

    (let* ((me (ormap is-me players))
	   (others (filter (lambda (p) (not (is-me p))) players))
	   (hashed-pos (players->hash-pos players))
	   (cache-entry (hash-mem? cache hashed-pos))
	   (cache-level (and cache-entry (car cache-entry)))
	   (cache-val (and cache-entry (cdr cache-entry))))


      (when (> eval-node-count eval-next-average)
	(set! eval-node-per-msec (/ (exact->inexact eval-node-count) (+ 1 (- (current-process-milliseconds)
									eval-time-started))))
	(set! eval-next-average (* 1.5 eval-next-average))
	;;(debug-f "node-per-msec: " eval-node-per-msec)
	)
      
      (when (= dept-to-go -1)
	(set! deadline-node-count (+ eval-node-count
				     (* (- deadline (current-process-milliseconds))
					eval-node-per-msec)))
	)


      ;;(debug-f "deadline-node-count: " dept-to-go (- deadline-node-count eval-node-count))
      (set! eval-node-count (+ eval-node-count 1))

      (cond ((not me) 
	     ;;(debug-f "dept " dept-to-go players " x ")
	     (make-val eval-max-val players eval-max-val players false))

	    ((eq? cache-level 'loop) 
	     ;;(debug-f "dept " dept-to-go players " L ")
	     false)

	    ((and cache-entry (>= (+ cache-level cache-dept-bonus) dept-to-go))
	     ;;(debug-f "dept " dept-to-go players " [" cache-level "] .")
	     cache-val)

	    ((or (< deadline (current-process-milliseconds))
		 (< deadline-node-count eval-node-count))
	     ;;(debug-f "dept " dept-to-go players " o")
	     ;;(debug-f "deadline-node-count: " dept-to-go deadline-node-count eval-node-count)	
	     ;;(debug-f "bf " eval-branch-factor)
	     (let ((rtn (eval-local board cell-map me players)))
	       (hash-put! cache hashed-pos (cons 0 rtn))
	       rtn))
	    (true
	     ;;(debug-f "dept " dept-to-go players " O")

	     (hash-put! cache hashed-pos (cons 'loop 'loop))
	     (let* ((branches 0)
		    (rtn (fold-the-future board me others
					  compound-me compound-adv compound-perm
					  compound-me-init compound-adv-init compound-perm-init
					;(* eval-perm-l-mult (- (length players) 1))
					  eval-perm-l
					  (lambda (pos)
					    (set! branches (+ branches 1))
					    (eval-may-recurse board cell-map my-id pos 
							      (- dept-to-go 1) cache deadline
							      (sub-deadline dept-to-go deadline-node-count
									    branches eval-branch-factor))))))

	       (hash-put! cache hashed-pos (cons dept-to-go rtn))
	       (set! eval-branch-factor (+ (* 0.2 eval-branch-factor)
					   (* 0.8 branches)))
	       rtn)))))
  

  (define (minimax-move board cell-map my-id players time-alloc cache)
    (val-my-move (minimax board cell-map my-id players time-alloc cache)))

  (define (minimax board cell-map my-id players time-alloc cache)
    (set! eval-node-count 0)
    (set! eval-next-average (* 10 eval-node-per-msec))
    (set! eval-time-started (current-process-milliseconds))
    (let-values (((rtn time b c)
		  (time-apply eval-may-recurse
			      (list board cell-map my-id players 0 cache
				    (exact->inexact (+ (current-process-milliseconds) time-alloc))
				    (* time-alloc eval-node-per-msec)))))
		;;(debug "-------------minimax time: " (/ time 1000.0) (format " going to ~a" (val-my-move (car rtn))))
		(car rtn)))
  

  (define (test)
    (define b (read-board-from-file "test.map"))
    (define pls (list (make-player 1 2 1 empty)
		      (make-player 2 2 2 empty)
		      (make-player 3 2 3 empty)
		      (make-player 4 2 4 empty)
		      ))
    
    ;;(debug-f (physics-apply-move-sequence b pls '(S W E S)))
    ;;(physics-apply-move! b pls (car pls) 'W)
    
    
    ;; (debug-f (player-possible-moves b (list-ref pls 1)))
    
    ;;(unzip (fold-all-moves b pls (lambda (mv acc) (cons mv acc)) empty))
    
    ;;(fold-permutations '(1 2 3 4 5 6 7 8) 10 (lambda (p acc) (debug "A " p) acc) false)
    
    (show-board b pls) (newline) (newline) (newline) 
    
    (let ((cache (make-hash 'equal)))
      (let ((rtn (minimax b (lambda (x y) 0) 3 pls 3000 cache)))
	(debug-f rtn)
	(show-board b (val-min-pls rtn)) (newline)
	(show-board b (val-max-pls rtn)) (newline))))

)
