(module pos-cache mzscheme
  
  (require "base-gm.ss" "physics.ss" "cqueue.ss")
  (provide make-history history-tostring record-position!)
  
  (define-struct t (max-size hash garbage-queue))
  (define-struct entry (hash-pos time-stamp used))
  
  (define (make-history max-size) (make-t max-size (make-hash 'equal) (make-cqueue)))
  
  (define (history-tostring his)
    (if (not (t? his)) false
	(format "<his: msz ~a ~a>" (t-max-size his) 
		(cqueue-map (lambda (e) (entry-tostring e)) (t-garbage-queue his)))))
  
  (define (entry-tostring entry)
    (if (not (entry? entry)) false
	(format "~a" (entry-hash-pos entry))))

  (define debug-f (make-debug player-tostring history-tostring cqueue-tostring entry-tostring))
  
  (define (throw-garbage! his)
    (let* ((h (t-hash his))
	   (q (t-garbage-queue his))
	   (f (cqueue-first q)))
      
      (if (not (entry-used f)) 
	  (begin (cqueue-remove-first! q)
		 (hash-remove! h f))
	  
	  (begin (set-entry-used! f false)
		 (cqueue-remove-first! q)
		 (cqueue-add-last! q f)
		 (throw-garbage! his)))))
  
  ;; Returns the time we last seen the position, or false
  (define (record-position! his time-stamp players)
    (let* ((hash-pos (players->hash-pos players))
	   (h (t-hash his))
	   (q (t-garbage-queue his))
	   (entry (hash-mem? h hash-pos)))
      
      (if (not entry)
	  (begin (if (= (cqueue-size q) (t-max-size his))
		     (throw-garbage! his))
		 (let ((new-entry (make-entry hash-pos time-stamp false)))
		   (hash-put! h hash-pos new-entry)
		   (cqueue-add-last! q new-entry))
		 false)
	  
	  (begin (set-entry-used! entry true)
		 (entry-time-stamp entry)))))
  
  (define (test)
    (define positions (map (lambda (i) (list (make-player 0 0 i empty))) '(1 2 3 4 5 6 7 8 9   4 5 6   10 11 12 13 14 15)))
    
    (define his (make-history 5))
    
    (debug-f "P" positions)
    
    (for-each (lambda (p)
		(debug-f "T" (record-position! his 0 p)  his))
	      positions))
)