
(module ptree mzscheme 
  (require "base-gm.ss" "heap.ss" (lib "match.ss"))

  (provide make-ptree ptree-insert ptree-find ptree-remove ptree-near ptree-tostring)
  
  (define-struct tree (size root))
  (define-struct node (x y left right) (make-inspector))
  (define-struct context (node parent verticaly))
  
  (define debug-f (make-debug ptree-tostring node-tostring-brief))
  
  (define (make-ptree) (make-tree 0 false))
  
  (define (ptree-node-leftof? verticaly a b)
    (ptree-leftof? verticaly (node-x a) (node-y a) b))
  
  (define (ptree-leftof? verticaly x y node)
    (or (and verticaly (< y (node-y node)))
	(and (not verticaly) (< x (node-x node)))))
  
  (define (ptree-insert tree x y)
    (if (not (tree-root tree))
	(set-tree-root! tree (make-node x y false false))
	
	(let loop ((node (tree-root tree))
		   (verticaly false))
	  (if (ptree-leftof? verticaly x y node)
	      (if (not (node-left node)) 
		  (set-node-left! node (make-node x y false false))
		  (loop (node-left node) (not verticaly)))
	      
	      (if (not (node-right node))
		  (set-node-right! node (make-node x y false false))
		  (loop (node-right node) (not verticaly))))))
    (set-tree-size! tree (+ (tree-size tree) 1)))
  
  (define (ptree-find tree x y)
    (if (not (tree-root tree)) false
	(let loop ((node (tree-root tree))
		   (parent false)
		   (verticaly false))
	  (cond ((and (= (node-x node) x)
		      (= (node-y node) y)) (make-context node parent verticaly))
		
		((ptree-leftof? verticaly x y node)
		 (if (not (node-left node)) false
		     (loop (node-left node) node (not verticaly))))
		
		(true 
		 (if (not (node-right node)) false
		     (loop (node-right node) node (not verticaly))))))))
  
  (define (ptree-find-best node parent verticaly want-leftmost want-verticaly)
    (define (best-of-three a b c)
      (foldl (lambda (x rtn) (cond ((not x) rtn)
				   ((not rtn) x)
				   ((eq? want-leftmost (ptree-node-leftof? want-verticaly x rtn)) x)
				   (true rtn)))
	     false (list a b c)))
    
    (cond ((not (eq? verticaly want-verticaly))
	   (let* ((a (and (node-left node) (ptree-find-best (node-left node) node (not verticaly)
							    want-leftmost want-verticaly)))
		  (b (and (node-right node) (ptree-find-best (node-right node) node (not verticaly)
							     want-leftmost want-verticaly)))
		  (rtn (best-of-three (and a (context-node a)) (and b (context-node b)) node)))
	     (if (eq? rtn node) (make-context node parent verticaly)
		 (make-context rtn node (not verticaly)))))
	  
	  ((and want-leftmost (node-left node)) (ptree-find-best (node-left node) node (not verticaly)
								 want-leftmost want-verticaly))
	  ((and (not want-leftmost) (node-right node)) (ptree-find-best (node-right node) node (not verticaly)
									want-leftmost want-verticaly))
	  
	  (true (make-context node parent verticaly))))
  
  
  (define (ptree-find-remplacement node parent verticaly)
    (cond ((and (not (node-left node)) (node-right node)) (make-context (node-right node) node (not verticaly)))
	  ((and (not (node-right node)) (node-left node)) (make-context (node-left node) node (not verticaly)))
	  (true (or (ptree-find-best (node-right node) node (not verticaly) true verticaly)
		    (ptree-find-best (node-left node) node (not verticaly) false verticaly)))))
  
  (define (ptree-remove-node tree node parent verticaly)
    (define (set-parent! n)
      (cond ((not parent) (set-tree-root! tree n))
	    ((eq? node (node-right parent)) (set-node-right! parent n))
	    (true (set-node-left! parent n))))
    

    (cond ((and (not (node-left node)) (not (node-right node)))
	   (set-parent! false))
	  ((and (node-left node) (not (node-right node)))
	   (set-parent! (node-left node)))
	  ((and (node-right node) (not (node-left node)))
	   (set-parent! (node-right node)))
	  (true

	   (let ((remp (ptree-find-remplacement node parent verticaly)))
	     (set-parent! (context-node remp))
	   
	     (ptree-remove-node tree (context-node remp) (context-parent remp) (context-verticaly remp))
	     (set-node-left! (context-node remp) (node-left node))
	     (set-node-right! (context-node remp) (node-right node))))))


  (define (ptree-remove tree x y)
    (let ((context (ptree-find tree x y)))
      (if (not context) (error "ptree-remove: not found")
	  (begin (set-tree-size! tree (- (tree-size tree) 1))
		 (ptree-remove-node tree (context-node context) 
				    (context-parent context) (context-verticaly context))))))


  (define (sqr x) (* x x))

  (define (distance node x y)
    (sqrt (+ (sqr (- (node-x node) x)) (sqr (- (node-y node) y))))) 

  (define (distance-1d verticaly node x y)
    (if verticaly (abs (- (node-y node) y))
	(abs (- (node-x node) x))))

  (define (ptree-near tree x y max-distance max-count)
    (define (dist node) (distance node x y))

    (define (near-far-subnode node verticaly)
      (if (ptree-leftof? verticaly x y node) (cons (node-left node) (node-right node))
	  (cons (node-right node) (node-left node))))

    (define (may-add q node cut-off)
      (let ((d (dist node)))
	(if (and cut-off (>= d cut-off)) empty
	    (begin (if (and max-count (= (heap-size q) max-count)) 
		       (begin
			 ;;(debug-f "Removing " (heap-peak q) (dist (heap-peak q)))
			 (heap-pop q)))
		   ;;(debug-f "Adding " node (dist node))
		   (heap-insert q node)))))

    (define (get-cut-off q)
      (cond ((and max-count (= (heap-size q) max-count)) (dist (heap-peak q)))
	    (max-distance max-distance)
	    (true false)))

    (let ((q (make-pqueue
	      (lambda (a b) (> (dist a) (dist b)))
	      (lambda (a b) (and (= (node-x a) (node-x b))
				 (= (node-y a) (node-y b)))))))

      (let loop ((node (tree-root tree))
		 (verticaly false))

	(if (not node) empty
	    (match-let (((near-subnode . far-subnode) (near-far-subnode node verticaly)))
		       (loop near-subnode (not verticaly))
		       (may-add q node (get-cut-off q))
		       (let ((cut-off (get-cut-off q)))
			 (if (or (not cut-off) (< (distance-1d verticaly node x y) cut-off))
			     (loop far-subnode (not verticaly)))))))
      
      (let loop ((rtn empty))
	(if (heap-empty? q) rtn
	    (loop (cons (heap-pop q) rtn))))))


  (define (ptree-tostring tree)
    (if (not (tree? tree)) false
	(string-append "ptree: sz " 
		       (number->string (tree-size tree))
		       "\n"
		       (node-tostring (tree-root tree)))))

  (define (node-tostring-brief node)
    (if (not (node? node)) false
	(string-append "<" (number->string (node-x node)) ", " (number->string (node-y node)) ">")))

  (define (node-tostring node)
    (if (not (node? node)) false
	(string-append (node-tostring-brief node) "\n"
		       (if (node-left node) (string-append "(" (node-tostring (node-left node)) ")") "nil") "\n"
		       (if (node-right node) (string-append "(" (node-tostring (node-right node)) ")") "nil") "\n")))
    

  
  (define (test)
    (define t (make-ptree))
    (for x 0 9
	 (for y 0 9 (ptree-insert t x y)))
    ;; (debug-f (ptree-tostring t))
  
    (debug-f (ptree-near t 5 5 7 10)))

  (test) (exit)
)
  