(module base mzscheme
  (require (lib "list.ss")
	   (lib  "etc.ss"))
  
  (provide cons-to-end
	   assoc-get
	   debug
	   to-string
	   member-eq?	   
	   string->char
	   last
	   member-str?
	   quicksort-vector
	   for
	   for-vector
	   but-last
	   halt
           enumerate
           for-list
           first-true
	   )

  (define (first-true x)
    true)
  
  
  (define-syntax for-list
    (syntax-rules
     (with)
     [(for-list ls with x body ...)
      (let ((x 'dummy))
        (letrec ((loop
                  (lambda (param)
                    (if (empty? param)
                        'done
                        (begin
                          (set! x (car param))
                          body ...
                          (loop (rest param)))))))
          (loop ls)))]))

  
  (define (enumerate a b)
    (cond
      ((> a b) empty)
      (else (cons a (enumerate (+ a 1) b)))))
  
  
(define-syntax halt
  (syntax-rules
      ()
    [(halt arg ...)
     (begin
       (debug "There was a problem: " arg ...)
       (error "Error."))]))  

;(define (halt . args)
;  (apply error
;	 (list (string-append
;		"Error: "
;		(foldl string-append "" (reverse (map to-string args)))))))

(define (but-last ls)
  (cond
   ((empty? ls) (error "incorrect list to butlast"))
   ((empty? (rest ls)) empty)
   (else (cons (first ls) (but-last (rest ls))))))

(define-syntax for
  (syntax-rules
   ()
   [(for x start stop body ...)
    (let ((x start))
      (letrec ((loop
                (lambda ()
                  (if (> x stop)
                      'done
                      (begin
                        body ...
                        (set! x (+ x 1))
                        (loop))))))
        (loop)))]))

(define-syntax for-vector
  (syntax-rules
      (with)
    [(for-vector v with x body ...)
     (for x 0 (- (vector-length v) 1)
	  body ...)]))

(define (quicksort-vector v less-than)
  (let ([count (vector-length v)])
    (let loop ([min 0][max count])
      (if (< min (sub1 max))
	  (let ([pval (vector-ref v min)])
	    (let pivot-loop ([pivot min]
			     [pos (add1 min)])
	      (if (< pos max)
		  (let ([cval (vector-ref v pos)])
		    (if (less-than cval pval)
			(begin
			  (vector-set! v pos (vector-ref v pivot))
			  (vector-set! v pivot cval)
			  (pivot-loop (add1 pivot) (add1 pos)))
			(pivot-loop pivot (add1 pos))))
		  (if (= min pivot)
		      (loop (add1 pivot) max)
		      (begin
			(loop min pivot)
			(loop pivot max)))))))))
  v)


	     
    

(define (member-str? s ls)
  (cond
   ((empty? ls) false)
   ((string=? s (first ls)) true)
   (else (member-str? s (rest ls)))))

(define (last ls)
  (cond
   ((empty? ls) (error "took a last but it was emptry"))
   ((empty? (rest ls)) (first ls))
   (else (last (rest ls)))))

(define (string->char s)
  (first (string->list s)))

(define (member-eq? x ls)
  (not (empty? (filter (lambda (y) (eq? x y)) ls))))

(define (to-string arg)
  (cond
   ((char? arg) (list->string (list arg)))
   ((string? arg) arg)
   ((symbol? arg) (symbol->string arg))
   ((number? arg) (number->string arg))
   ((list? arg) (string-append
		 "("
		 (foldl string-append "" (map to-string arg))
		 ")"))
   (else (halt "to-string: " arg))))

(define (debug . args)
 'done)
  
  ; (for-each display args)
 ; (newline))

(define (assoc-get label ls)
  (cond
    ((empty? ls) (error (string-append "failed to find " (to-string label))))
    ((eq? label (first (first ls)))
     (first ls))
    (else (assoc-get label (rest ls)))))

(define (cons-to-end a ls)
  (cond
   ((empty? ls) (cons a ls))
   (else (cons (first ls)
               (cons-to-end a (rest ls))))))


)