;; THE BRAIN INTERFACES

(require "../dbt/client.ss"
         "../dbt/board.ss"
         "../dbt/types.ss"
         "../dbt/viz.ss"
         (lib "list.ss")
         "../rob/base.ss"
         (lib "posn.ss" "lang"))

(load "../rob/search.scm")

;; Type: info
;; Comment: info is the main info structure.
;;          Should be everything that's important about the current state.
;; Interface:
;;  command, arg* -> result
;; Specifics:
;;  (info 'x) -> your x position
;;  (info 'y) -> my y position


;; Type: path-planner
;; Interface:
;;  info, to_x, to_y -> direction

;; Type: bidder
;; Interface:
;;  info, action -> bid amount
;;    bid amount is an integer


;; Type: goal-generator
;; Interface:
;;  info -> place

;; Type: stops
;; Interface:
;;  (get-closest stops)
;;  ;; checks to make sure the return location has packages
;;  ;;   or is a base.

;; Type: place (place is now posn)

;; tools to create clients in a plug and play fashion
(define-struct client (init round info))

(define board-viz #f)

(define-struct extra (init pieces))

(define default-extra
  (make-extra (lambda (board id capacity) 'done) empty))

(define (create-client path-planner bidder goal-generator stops an-extra)
  
  (define result (make-client 'dummy 'dummy 'dummy))
  
  (define pre-info 'dummy)
  
  (define my-viz-flag false)
  
  (define extra-result 'dummy)
  
  (define (init-handler board my-id capacity)
    (set! pre-info (make-info board my-id capacity))
    
    ;;initialize the stopper function
    ((stops-init stops) board stops)
    
    ;;initialize the extra ...it better be a thunk
    (set! extra-result ((extra-init an-extra) board my-id capacity))
    
    ;;Dave's Viz
    ;(if (and (not (empty? viz-flag)) (first viz-flag))
    ;    (begin
          ;(set! my-viz-flag true)
          (set! board-viz (instantiate viz% () (board board) (my-id my-id)))
          (send board-viz draw-board '() '() 0)
          (sleep/yield 1)
    
    (void))
  

  (define (usps-package-tracking! all-packages package-data)
    (debug "all-packages is " all-packages))
  
  (define (round-handler money-left players all-packages here-packages updates my-score package-data)
    (debug "places are " (stops-places stops))
    
    (usps-package-tracking! all-packages package-data)
    
    ;;(places-to-go package-data all-packages)
    
    (set-client-info! result (pre-info money-left players all-packages here-packages (extra-pieces an-extra)))
    
    ;;board viz
    ;(if my-viz-flag
     ;   (begin
          (send board-viz draw-board players updates my-score)
          (sleep/yield 1)
    
    
    (debug "you are at position " ((client-info result) 'x) ","  ((client-info result) 'y))
    
    (let* ((goal (goal-generator (client-info result) stops))
           (action (path-planner (client-info result) (posn-x goal) (posn-y goal)))
           (bid (bidder (client-info result) action)))
    
      (make-command bid action)))
  
  (set-client-init! result init-handler)
  (set-client-round! result round-handler)
  
  result ;;return the new client
  )


;;simple (trivial) versions of these types


;; info X action -> bid
(define (simple-bidder info action)
  1)


;; info X tox X toy -> action
;;  this pretends there are no water or walls or other robots
(define (simple-path-planner info tox toy)
  (debug "to x is " tox " and to y is " toy)
   (cond
     [(info 'packages?) (pick-up-packages info)]
     [(> (info 'x) tox) (make-move 'W)]
     [(< (info 'x) tox) (make-move 'E)]
     [(> (info 'y) toy) (make-move 'S)]
     [(< (info 'y) toy) (make-move 'N)]
     [(and (= (info 'x) tox)
           (= (info 'y) toy))
      (drop-packages info)]
     [else (halt "your path planner doesn't know what to do.")]))


(define (drop-packages info)
  (debug "packages are " (info 'packages))
  ;; you want to make sure that you don't drop packages 
  ;; when you don't have any
  ;; (unless you really want a no op)
  ;; this is a waste.
  (make-drop 
   (map package-id
        (filter (lambda (p)
                  (and (= (package-dest-x p) (info 'x))
                       (= (package-dest-y p) (info 'y))))
                (ids->pkgs (info 'my-package-ids) info)))))

(define (ids->pkgs ids info)
  (map
   (lambda (id)
     (find-package id (info 'all-packages)))
   ids))

;; info -> action
(define (pick-up-packages info)
  (debug "you are picking up packages")
  (make-pick (choose-packages info)))

;; info -> list(package)
;; this chooses packages using the knapsack solver.
(define (choose-packages info)
  (debug "packages here are " (info 'packages))
  (map
   item-id
   (knapsack-list (map (lambda (p)
                         (make-item (package-id p) (package-weight p)))
                       (info 'packages))
                  (info 'sack-size))))

;; info X positional goal struct
(define (simple-goals info stops)
  (debug "in simple goals; stops places is " (stops-places stops))
  ((stops-update! stops) stops info)
  (if (info 'deliveries?)
      (begin
        (debug "there are deliveries")
        (let ((next-pkg (first (info 'deliveries))))
          (make-place
           (info 'dest-x next-pkg)
           (info 'dest-y next-pkg)))
        )
      ((stops-nearest stops) stops)))
      

(define-struct stops (places init nearest update!))


;; grabs the first, or, if empty, goes to <1,1>
(define (simple-nearest stops)
  (if (empty? (stops-places stops))
      (make-posn 1 1)
      (first (stops-places stops))))
         

(define (simple-update! stops info)
  ;;this should update only if necessary
  (let ((x (info 'x))
        (y (info 'y)))
    (set-stops-places! stops
                       (let loop ((ls (stops-places stops)))
                         (cond
                           ((empty? ls) empty)
                           ((and (= x (posn-x (first ls)))
                                 (= y (posn-y (first ls))))
                            (rest ls))
                           (else (cons (first ls) (loop (rest ls)))))))))
                

;;a struct
(define simple-stops
  (make-stops empty 
              (lambda (board stops)
                (set-stops-places! stops (find-all-bases board)))
              simple-nearest
              simple-update!))

;; find all the bases on a board
(define (find-all-bases board)
  (define acc empty)
  
  (for down 1 (board-height board)
    (for across 1 (board-width board)
         (if (eq? (board-ref board across down) 'home)
             (set! acc (cons (make-posn across down)
                             acc)))))
  acc)



;; INFO
(define (make-info board my-id capacity)
  (lambda (money-left players all-packages here-packages extra-pieces)
    (for-each (lambda (p)
                (add-package-hash! (package-id p) p))
              all-packages)
    (lambda (com . args)            
      (cond
        [(eq? com 'x) (player-x (find-player my-id players))]
        [(eq? com 'y) (player-y (find-player my-id players))]
        [(eq? com 'packages?) (not (empty? here-packages))]
        [(eq? com 'packages) here-packages]
        [(eq? com 'can-take-more-here?)
         (let ((smallest (apply min (map package-weight here-packages))))
           (debug "smallest package is " smallest "; capacity is " capacity)
           (>= (- capacity (my-space-used (find-player my-id players))) smallest))]
        [(eq? com 'all-packages) all-packages]
        [(eq? com 'package-ids) (map (lambda (p) (package-id p)) here-packages)]
        [(eq? com 'my-package-ids) (player-package-ids (find-player my-id players))]
        [(eq? com 'sack-size) capacity]
        [(eq? com 'deliveries?) (not (empty? (player-package-ids (find-player my-id players))))]
        [(eq? com 'deliveries) (player-package-ids (find-player my-id players))] ;;returns ids
        [(eq? com 'dest-x) (package-dest-x (find-package (first args) all-packages))]
        [(eq? com 'dest-y) (package-dest-y (find-package (first args) all-packages))]
        [(eq? com 'extra) extra-pieces]
        [(eq? com 'board) board]
        [(eq? com 'dest-posn) (make-posn (package-dest-x (find-package
                                                          (first args)
                                                          all-packages)) ;;needs to ids->packages
                                         (package-dest-y (find-package
                                                          (first args)
                                                          all-packages)))]
        [else (halt "You passed a command to info which was not understood.  Command=" com)]))))

(define (find-player id players)
  (cond
    [(empty? players) (halt "I was searching for id " id " but it doesn't exist in players.")]
    [(= id (player-id (first players)))
     (first players)]
    [else (find-player id (rest players))]))

(define (find-package id packages)
  (cond
    [(empty? packages) (halt "I was searching for id " id " but it doesn't exist in packages.")]
    [(= id (package-id (first packages)))
     (first packages)]
    [else (find-package id (rest packages))]))

(define (posn=? a b)
  (and (= (posn-x a)
          (posn-x b))
       (= (posn-y a)
          (posn-y b))))
