;; THE BRAIN INTERFACES

(require "../dbt/client.ss"
         "../dbt/board.ss"
         "../dbt/types.ss"
         "../dbt/viz.ss"
         (lib "list.ss")
         "../rob/base.ss"
         (lib "posn.ss" "lang")
         "../rob/tsp-rob.ss" )

(load "../rob/search.scm")

;; Type: info
;; Comment: info is the main info structure.
;;          Should be everything that's important about the current state.
;; Interface:
;;  command, arg* -> result
;; Specifics:
;;  (info 'x) -> your x position
;;  (info 'y) -> my y position


;; Type: path-planner
;; Interface:
;;  info, to_x, to_y -> direction

;; Type: bidder
;; Interface:
;;  info, action -> bid amount
;;    bid amount is an integer


;; Type: goal-generator
;; Interface:
;;  info -> place

;; Type: stops
;; Interface:
;;  (get-closest stops)
;;  ;; checks to make sure the return location has packages
;;  ;;   or is a base.

;; Type: place (place is now posn)

;; tools to create clients in a plug and play fashion
(define-struct client (init round info))

(define-struct me-hack (myself packages))

(define me-me (make-me-hack 'dummy 'dummy))

(define board-viz #f)


(define the-bidder 'dummy-bidder)

(define global-direction-f 'dummy)
(define global-distance-f 'dummy)


;; GLOBAL VAR
(define global-tsp-count -1)
(define TSP-FREQ-PARAM 10)

(define-struct extra (init pieces))

(define default-extra
  (make-extra (lambda (board id capacity) 'done) empty))

(define (create-client path-planner bidder goal-generator stops an-extra)
  
  (define result (make-client 'dummy 'dummy 'dummy))
  
  (define pre-info 'dummy)
  
  (define my-viz-flag false)
  
  (define extra-result 'dummy)
  
  (define (init-handler board my-id capacity)
    (set! pre-info (make-info board my-id capacity))
    
    ;;initialize the stopper function
    ((stops-init stops) board stops)
    
    ;;initialize the extra ...it better be a thunk
    (set! extra-result ((extra-init an-extra) board my-id capacity))
    
    ;;Dave's Viz
    ;;(set! my-viz-flag true)
    ;;(set! board-viz (instantiate viz% () (board board) (my-id my-id)))
    ;;(send board-viz draw-board '() '() 0)
    ;;(sleep/yield 1)
    
    (void))
  
  
  ;; add on your deliveries
 (define (to-deliver mine-packages)
   (debug "mine packages is " mine-packages)
   (map (lambda (pkg)
          (make-posn 
           (package-dest-x pkg)
           (package-dest-y pkg)))
        mine-packages))
  
  
    
    
  
  (define (round-handler money-left players all-packages here-packages updates my-score package-data
                         possible-package-posns whoami)
    
    (debug "plain stops is " stops)
    
    (set-me-hack-myself! me-me whoami)
    (set-me-hack-packages! me-me (player-packages (player-id whoami) players all-packages))
    (debug "just set packages " (me-hack-packages me-me))
    
    (set-client-info! result (pre-info money-left players all-packages here-packages (extra-pieces an-extra)))
    
    ;;update your goals here.
    (set! global-tsp-count (+ 1 global-tsp-count))
    
    (if (= 0 (modulo global-tsp-count TSP-FREQ-PARAM))
        (begin
          (debug "to-deliver is " (to-deliver (me-hack-packages me-me)))
           (let-values (((posn dist-f dir-f)
                         (get-next-goal ((client-info result) 'board)
                                        (make-posn ((client-info result) 'x) ((client-info result) 'y))
                                        (append (to-deliver (me-hack-packages me-me)) possible-package-posns))))
             (set-stops-places! stops posn)
             (set! global-direction-f dir-f)
             (set! global-distance-f dist-f))))
     
     ;;(set-stops-places! stops possible-package-posns)
    
    
    (debug "places are " (stops-places stops))
    
    
    
    ;;(places-to-go package-data all-packages)
    
    
    
    ;;(usps-package-tracking! package-data (client-info result))
    
    
    ;; GUI FUNCTIONS
   ;; (send board-viz draw-board players updates my-score)
   ;; (sleep/yield 1)
    
    
    (debug "you are at position " ((client-info result) 'x) ","  ((client-info result) 'y))
    
    (let* ((goal (goal-generator (client-info result) stops))
           (dummy (debug "hey rob, your goal is " goal))
           (dummy2 (if (eq? goal 'dummy) (set! goal (make-posn ((client-info result) 'x)
                                                               ((client-info result) 'y)))))
           (action (path-planner (client-info result) (posn-x goal) (posn-y goal)))
           (bid (bidder (client-info result) action)))
    
      (make-command bid action)))
  
  (set-client-init! result init-handler)
  (set-client-round! result round-handler)
  
  result ;;return the new client
  )


;;simple (trivial) versions of these types


(define percent-bid-pick-param .01)
(define percent-bid-drop-param .02)

;; info X action -> bid
(define (simple-bidder info action)
 1)
 ; (cond
 ;   ((drop? action) (+ 3 (floor (* percent-bid-drop-param 100))))
 ;   ((drop? action) (+ 3 (floor (* percent-bid-drop-param 100))))
 ;   (else (let ((r (random 10)))
 ;           (if (<= r 8)
 ;               1
 ;               (+ 1 (random 4)))))))
     


;; info X tox X toy -> action
;;  this pretends there are no water or walls or other robots
(define (simple-path-planner info tox toy)
  (debug "to x is " tox " and to y is " toy)
   (cond
     [(info 'packages?) (pick-up-packages info)]
     [(> (info 'x) tox) (make-move 'W)]
     [(< (info 'x) tox) (make-move 'E)]
     [(> (info 'y) toy) (make-move 'S)]
     [(< (info 'y) toy) (make-move 'N)]
     [(and (= (info 'x) tox)
           (= (info 'y) toy))
      (drop-packages info)]
     [else (halt "your path planner doesn't know what to do.")]))


(define (drop-packages info)
  (debug "packages are " (info 'packages))
  ;; you want to make sure that you don't drop packages 
  ;; when you don't have any
  ;; (unless you really want a no op)
  ;; this is a waste.
  (make-drop 
   (map package-id
        (filter (lambda (p)
                  (and (= (package-dest-x p) (info 'x))
                       (= (package-dest-y p) (info 'y))))
                (ids->pkgs (info 'my-package-ids) info)))))

(define (ids->pkgs ids info)
  (map
   (lambda (id)
     (find-package id (info 'all-packages)))
   ids))

;; info -> action
(define (pick-up-packages info)
  (debug "you are picking up packages")
  (make-pick (choose-packages info)))

;; info -> list(package)
;; this chooses packages using the knapsack solver.
(define (choose-packages info)
  (debug "packages here are " (info 'packages))
  (map
   item-id
   (knapsack-list (map (lambda (p)
                         (make-item (package-id p) (package-weight p)))
                       (info 'packages))
                  (info 'sack-size))))

;; info X positional goal struct
(define (simple-goals info stops)
  (debug "in simple goals; stops places is " (stops-places stops))
  ((stops-update! stops) stops info)
  (if (info 'deliveries?)
      (begin
        (debug "there are deliveries")
        (let ((next-pkg (first (info 'deliveries))))
          (make-place
           (info 'dest-x next-pkg)
           (info 'dest-y next-pkg)))
        )
      ((stops-nearest stops) stops)))
      

(define-struct stops (places init nearest update!))


;; grabs the first, or, if empty, goes to <1,1>
(define (simple-nearest stops)
  (debug "in simple nearest with " (stops-places stops))
  (if (empty? (stops-places stops))
      (make-posn 1 1)
      (first (stops-places stops))))
         

(define (simple-update! stops info)
  ;;this should update only if necessary
  (let ((x (info 'x))
        (y (info 'y)))
    (set-stops-places! stops
                       (let loop ((ls (stops-places stops)))
                         (cond
                           ((empty? ls) empty)
                           ((and (= x (posn-x (first ls)))
                                 (= y (posn-y (first ls))))
                            (rest ls))
                           (else (cons (first ls) (loop (rest ls)))))))))
                

;;a struct
(define simple-stops
  (make-stops empty 
              (lambda (board stops)
                (set-stops-places! stops (find-all-bases board)))
              simple-nearest
              simple-update!))

;; find all the bases on a board
;(define (find-all-bases board)
;  (define acc empty)
;  
;  (for down 1 (board-height board)
;    (for across 1 (board-width board)
;         (if (eq? (board-ref board across down) 'home)
;             (set! acc (cons (make-posn across down)
;                             acc)))))
;  acc)



(define (my-weight)
  (foldl + 0 (map package-weight (me-hack-packages me-me))))

;; INFO
(define (make-info board my-id capacity)
  (lambda (money-left players all-packages here-packages extra-pieces)
    ;(for-each (lambda (p)
    ;            (add-package-hash! (package-id p) p))
    ;          all-packages)
    (lambda (com . args)            
      (cond
        [(eq? com 'x) (player-x (find-player my-id players))]
        [(eq? com 'y) (player-y (find-player my-id players))]
        [(eq? com 'packages?) (not (empty? here-packages))]
        [(eq? com 'packages) here-packages]
        [(eq? com 'at-base?) (eq? 'home (board-ref board
                                                   (player-x (find-player my-id players))
                                                   (player-y (find-player my-id players))))]
        [(eq? com 'can-take-more-here?)
         (let ((smallest (apply min (map package-weight here-packages))))
           (debug "smallest package is " smallest "; capacity is " capacity)
           (>= (- capacity (my-weight)) smallest))]
        [(eq? com 'all-packages) all-packages]
        [(eq? com 'package-ids) (map (lambda (p) (package-id p)) here-packages)]
        [(eq? com 'my-id) my-id]
        [(eq? com 'players) players]
        [(eq? com 'my-package-ids) (player-package-ids (find-player my-id players))]
        [(eq? com 'sack-size) capacity]
        [(eq? com 'deliveries?) (not (empty? (player-package-ids (find-player my-id players))))]
        [(eq? com 'deliveries) (player-package-ids (find-player my-id players))] ;;returns ids
        [(eq? com 'dest-x) (package-dest-x (find-package (first args) all-packages))]
        [(eq? com 'dest-y) (package-dest-y (find-package (first args) all-packages))]
        [(eq? com 'extra) extra-pieces]
        [(eq? com 'board) board]
        [(eq? com 'dest-posn) (make-posn (package-dest-x (find-package
                                                          (first args)
                                                          all-packages)) ;;needs to ids->packages
                                         (package-dest-y (find-package
                                                          (first args)
                                                          all-packages)))]
        [else (halt "You passed a command to info which was not understood.  Command=" com)]))))

(define (find-player id players)
  (cond
    [(empty? players) (halt "I was searching for id " id " but it doesn't exist in players.")]
    [(= id (player-id (first players)))
     (first players)]
    [else (find-player id (rest players))]))

(define (find-package id packages)
  (cond
    [(empty? packages) (halt "I was searching for id " id " but it doesn't exist in packages.")]
    [(= id (package-id (first packages)))
     (first packages)]
    [else (find-package id (rest packages))]))

(define (posn=? a b)
  (and (= (posn-x a)
          (posn-x b))
       (= (posn-y a)
          (posn-y b))))

(define live-packages empty)


;; USPS

(define (usps-package-tracking! package-data info)
  ;; if we are at a base, add the packages that are there to live-packages
  (if (info 'at-base?)
      (set! live-packages (merge-packages (info 'packages) live-packages)))
  ;(debug "live packages is *first*  " (map package-id live-packages))
  ;; now check the changes reflected in package-data
  (set! live-packages 
        (filter
         (lambda (p)
           (not (package-done? p (filter (lambda (x) (= (length x) 3)) package-data))))
         live-packages))
  
  (debug "live packages is " (map package-id live-packages))
  )


(define (package-done? pkg data-ls)
  (cond
    ((empty? data-ls) false)
    (else (let* ((id (first (first data-ls)))
                 (cur-pkg (id->package id)))
            (if (and
                 (eq? pkg cur-pkg)
                 (= (package-dest-x pkg) (second (first data-ls)))
                 (= (package-dest-y pkg) (third (first data-ls))))
                true
                (package-done? pkg (rest data-ls)))))))

(define (merge-packages a b)
  (cond
    ((empty? a) b)
    ((member-pkg? (first a) b) (merge-packages (rest a) b))
    (else (cons (first a) (merge-packages (rest a) b)))))

(define (member-pkg? pkg ls)
  (cond
    ((empty? ls) false)
    ((= (package-id pkg) (package-id (first ls))))
    (else (member-pkg? pkg (rest ls)))))
