;;KNAPSACK


(require (lib "list.ss")
         (lib "etc.ss")
         (file "../rob/base.ss"))


(define-struct item (id weight))
(define (item->string i)
  (string-append "<"
                 (to-string (item-id i))
                 ","
                 (to-string (item-weight i))
                 ">"))

(define (knapsack-list item-ls sack-size)
  (debug "KNAPSACK: item ls size is: " (length item-ls))
  (let ((ls (sort-items item-ls)))
    (set! best-ls (greedy ls sack-size))
    (debug "best-ls is " best-ls)
    (set! best (foldl + 0 (map item-weight best-ls)))
    (set! to-do (cons
                 (make-state
                  empty
                  ls
                  sack-size
                  0)
                 to-do))
    (search)))

(define (compare-items a b)
  (> (item-weight a) (item-weight b)))

;;recommends a (descending) sorted list
(define (greedy item-ls sack-size)
  (cond
    ((empty? item-ls) empty)
    ((> (item-weight (first item-ls))
        sack-size)
     empty)
    (else (cond (first item-ls)
                (greedy (rest item-ls)
                        (- sack-size (item-weight item-ls)))))))

;; does NOT mutate
(define (sort-items item-ls)
  (quicksort item-ls compare-items))


;;we need a lowerbound to prune from
;; this will give a fast way to compute
;; a reasonable knapsack solution
;; prune as soon as it can't beat it.
;; Remaining-items is a list of items
;; LOWERBOUND IS PROBABLY NOT NECESSARY
;(define (lower-bound remaining-items-ls sack-space)
;  (cond
;    ((< sack-space 0) (halt "a rob logic error: ran out of sack space"))
;    ((= sack-space 0) 0)
;    ((> (item-weight (first remaining-items-ls))
;        sack-space)
;     0)
;    (else (+ (item-weight (first remaining-items-ls))
;             (lower-bound (rest remaining-items-ls)
;                          (- sack-space (item-weight (first remaining-items-ls)))
;                          )))))


;;trivial to upper-bound
;; am I doing this right?
(define (upper-bound st)
  (if (done? st)
      (state-so-far st)
      (+ (state-so-far st) (state-space st))))


;;branch and bound


;;a state in branch and bound
(define-struct state (fixed rest space so-far))

;;states to examine
(define to-do empty)
;; need a cons-to-end fast
(define best 0)
(define best-ls 'dummy)


(define (search)
  (search/stop 400)) ;;parameter tune.


;;the main search wrapper
;; should be given a sorted list (optimization)
(define (search/stop iter-stopper)
  ;;(debug "best-ls is " best-ls)
  (cond
    [(empty? to-do) best-ls]
    [(<= iter-stopper 0) (begin
                           (debug "iter stopping in effect with is= " iter-stopper "; best=" best-ls)
                           best-ls)]
    [else (begin
            (bb!)
            ;;(for-each state-print to-do)
            (search/stop (- iter-stopper 1)))]))
  
  
  
  
  
  
;; branch and bound function (bb!)
(define (bb!)
  (let ((prune (upper-bound (first to-do))))
    (cond 
      [(>= best prune) (begin
                         ;;(debug "you have pruned!")
                         (set! to-do (rest to-do)))]
      [(done? (first to-do)) (update-best!)]
      [else (expand!)])))

;;have we reached a leaf node in the search?
(define (done? st)
  (empty? (state-rest st)))

;;update-best!
(define (update-best!)
  (set! best (state-so-far (first to-do)))
  (set! best-ls (state-fixed (first to-do)))
  (set! to-do (rest to-do)))


;;(define-struct state (fixed rest space so-far))


(define (expand!)
  ;;expands the first and then removes it
  (let* ((st (first to-do))
         (next-item (first (state-rest st)))
         (on (make-state
              (cons next-item
                    (state-fixed st))
              (rest (state-rest st))
              (- (state-space st) (item-weight next-item))
              (+ (state-so-far st) (item-weight next-item))))
         (off (make-state
               (state-fixed st)
               (rest (state-rest st))
               (state-space st)
               (state-so-far st))))
    ;;need to check that on is 
    (set! to-do (append (rest to-do)
                        (keep-feasible (list on off))))))

(define (keep-feasible state-ls)
  (filter (lambda (st)
            (>= (state-space st) 0))
          state-ls))
              
  

(define (state-print st)
  (debug "fixed: " (apply string-append (map item->string (state-fixed st)))
         "; rest: " (apply string-append (map item->string (state-rest st)))
         "; space: " (state-space st)
         "; so-far: " (state-so-far st)))


(define-syntax item
  (syntax-rules 
   ()
   [(item id weight)
    (make-item id weight)]))
                

;;testing
(define base
  (list
   (item 0 10)
   (item 1 50)
   (item 2 100)
   (item 3 6)
   (item 4 7)))

(knapsack-list base 165)

